/* =============================================================================
 UPC - EETAC - CSD - http://digsys.cat

 Project P12 Tutorial 1: A fixed-time timer: FSM + TMR0
 
Common features:
- C language, coding in FSM style, dedicated processor approach, PICDEM2+ board 
- Interrupt-driven Trigger signal (a pushbutton to be connected at RB0/INT0)
- Two modes of operation: Retriggerable (RT=1) (see for example the classic chip 
 74LS122/123) and non-retriggerable (RT=0) (see the chip 74LS121).  
- RT value by polling the switch at RA4
- TMR0 peripheral to count real-time up to 18.5 s
- Compiler XC8, Microcontroller PIC16F877A
- Extra: A buzzer will sound for 2 s at 1 kHz once the timing period has ended. 

Approach: 
- Step 1: Study the notes available on the web about the design of a timer using
          VHDL and FPGA/CPLD devices. Also the P7 on a programmable timer. 
            - The hardware counter (DATAPATH) is now the TMR0 peripheral
            - The control Unit is now the software in FSM running the uC 
            - Draw a state diagram
    	  
Step 2: Prepare the hardware, so that adapt the PICDEM2+ board 
    - Timer_Out is a LED connected at RB3
    - Retriggerable signal (RT), the switch at RA4
    - The Trigger signal is RB0/INT0
    - Add a Buzzer if you like to hear a sound wave while timing. 		
    
Step 3: Program the init_system() function, configuring ports, interrupts 
        and the Timer0. 

---> For instance, in order to generate a timing period of 18.5 s : 
    FOSC/4 = 1MHz, PRESCALER = 1/16, TMR0 = (256- 250), a software variable 
    TMR0_POST_Scaler = 4625  (int);
              18.5 s = Time_base * N1 * N2 * N3 = 1 us * 16 * 250 * 4625          
          
    This is a TMR0 interrupt every 4 ms, which is going to be used to increment 
    the TMR0_POST_Scaler variable.
              
From the datasheet: 
T0CON: TIMER0 CONTROL REGISTER:

    T0CS: Timer0 Clock Source Select bit ==> 0 
            = Internal instruction cycle clock (CLKO)  [FOSC/4 = 1MHz]
    PSA: Timer0 Prescaler Assignment bit ==> 0 
            = Timer0 prescaler is assigned.
            Timer0 clock input comes from prescaler output.

    PS<2:0>: TMR0 PRE_Scaler select bits ==> 011 = 1:16 Prescale value

    TMR0_Value = 250 ==> TMR0 = (256 - TMR0_Value) (interrupt TMR0IF after 4 ms)

And, finally, the extra variable needed to suit the large timer period (N3). 
It will be updated by software at every TMR0 interrupt until 
the "TMR0_POST_Scaler_Max_Value" is reached:
    ==> TRM0_POST_Scaler_Max_Value = 4625 and so setting the
   software flag "Timer_Flag" that will advance the FSM
============================================================================= */

#include <xc.h>
   /* This general header file includes the device specific headers like
    <pic16f877a.h>. This file contains all the microcontroller
    declarations. All the header files are located in the C compilers's specific
    of: C:\Program Files (x86)\Microchip\xc8\v1.33\include\pic16f877a.h */

/* ------------------------------------
	Definitions
   -------------------------------------------------------------------------- */
/* The enumeration of all the machine states as a "Char" variable
   containing an ASCII character will simplify the debugging.
   "current_state" will contain only a valid state on the list: */

#define 	Idle	 	'A'
#define 	Start_TMR0	'B'
#define 	Timing 		'C'
#define 	Stop_TMR0	'D'

/* ------------------------------------
	Function prototypes
   -------------------------------------------------------------------------- */
// The three basic functions to organise the code to emulate a FSM
void init_system (void);
char state_logic(void);
char output_logic(void);

//The interrupt service routine (ISR) to detect the Trigger (external INT) or
// the TMR0 interrupt

static void interrupt service_routine (void);

// The auxiliary functions to read (poll) and write port pins and be able to
// organise the software variables
void write_outputs (void);
void read_inputs(void);

/* -----------------------------------
	Global variables
   -------------------------------------------------------------------------- */
/* Let's use convenient byte-type (char) variables to arganise the RAM. */

char Timer_Out_value; // Output of the Timer. "High" when timing
char Buzzer_value;    // Enhancement: Let's make a sound for 2 s 
                      // when the timing period ends. 

char current_state = Idle; /* The state variable: a RAM memory position which
                              will be updated ("asynchronously")
                              every time that "state_logic()" is executed */

char Trigger_Flag;	// The flag to start the timing operation

char RT_value;      // Retriggerable or non-retriggerable input

char TMR0_Flag;     // Software variable to handle the timng period from TMR0

/*To fix a time of 18.5 s*/
unsigned int TMR0_POST_Scaler_Max_Value;
unsigned int TMR0_POST_Scaler;
char TMR0_value;
				

/* =============================================================================
		Main function
   ========================================================================== */
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
	if (state_logic()) break;
	if (output_logic()) break;
    }
    while (1){
        __nop();// the system has crashed in an ilegal state and needs attention
               // This second infinite loop is never reached in normal operation
    }
    
}

/* ------------------------------------
	Function definitions
   -------------------------------------------------------------------------- */

/* =============================================================================
	Init the microcontroller
   ========================================================================== */
void init_system (void){

 /* These are the PIC16F877A configuration bits translated to the XC8 Compiler.
    They can be generated automatically in the MPLAB X IDE, but normally,
    you don't have to change them.
    So, copy and paste them as they are to your projects */

//#pragma config FOSC = EXTRC   // Oscillator Selection bits (RC oscillator)
#pragma config FOSC = XT     	// External XTAL Oscillator

//#pragma config WDTE = ON      // Watchdog Timer Enable bit (WDT enabled)
#pragma config WDTE = OFF       // Watchdog Timer not in use (WDT disabled)
//#pragma config PWRTE = OFF    // Power-up Timer Enable bit (PWRT disabled)

#pragma config PWRTE = ON      	// Power-up Timer Enable bit (PWRT enabled)

#pragma config BOREN = ON       // Brown-out Reset Enable bit (BOR enabled)
#pragma config LVP = ON         // Low-Voltage (Single-Supply) In-Circuit Serial
				// Programming Enable bit (RB3/PGM pin has PGM
				// function; low-voltage programming enabled)

#pragma config CPD = OFF        // Data EEPROM Memory Code Protection bit
        			//(Data EEPROM code protection off)
#pragma config WRT = OFF        // Flash Program Memory Write Enable bits (Write
				// protection off; all program memory may be
				// written to by EECON control)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit
                		// (Code protection off)

// Inicialise operations starts here:
    /* Pay attention to how the ports are configured
       using mainly PORTx and TRISx registers*/

    PORTA 	= 0x00; // Reset all Flip-Flops at PORTA
    ADCON1 	= 0x06;	/* All the port digital I/O.
            Remember that the PORTA can be also defined as analogue inputs*/
/*    ADCON1 	= 0x04;	 //3 analogue inputs and the others digital I/O.    */
    
    TRISA 	= 0b000010000;	// PORTA4 input RT

    PORTB 	= 0x00; 	// Reset all Flip-Flops at PORTB
    TRISB 	= 0b00000001;   /* RB0_INT is the external trigger input 
                                and the RB3 the Timer OUT */

    PORTC 	= 0x00; 	// The RC2 is the Buzzer output
    TRISC 	= 0b000000000;

    INTEDG	= 0;	// INT edge selection: '0' --> on falling edge
                        // (interrupt after clicking the pushbutton)
    INTE   = 1;	// Enable INT interrupts 
    
   // Let's init the state machine:
    current_state = Idle;
    Trigger_Flag = 0;
    TMR0_Flag = 0;
 
    // Let's init the Timer outputs:
    Timer_Out_value = 0;
    Buzzer_value = 0;

// Let's start the TMR0 to count FOSC/4 pulses (4MHz/4 = 1 MHz)
    T0CS = 0;
    TMR0IE = 0; /* Timer0 interrupts not allowed when initiating the
                microcontroller. TMR0 interrupts will be served when timing */


    GIE = 1;	// Global interrupts allowed
}
     


//*****************************************************************************
//State variable logic routine 
//*****************************************************************************
char state_logic(void){
    // local variable to check tha the FSM is always in a known state
    char error = 0;
    
    read_inputs();  // Sample the level sensitive inputs

    switch (current_state) {
        case Idle:
            if (Trigger_Flag == 1){
                Trigger_Flag = 0;
                current_state = Start_TMR0;
            }
            else 
                current_state = Idle;
            break;
        case Start_TMR0:
                current_state = Timing;
            break;
        case Timing:
            if (TMR0_Flag == 1 ){   // End of the timing period
                current_state = Stop_TMR0;
                TMR0_Flag = 0;
            }
            else if (Trigger_Flag == 1 && RT_value == 1){// retrigger the TMR0
                current_state = Start_TMR0;
                Trigger_Flag = 0;
            }
            else
                current_state = Timing;
            break;
			
        case Stop_TMR0:
            current_state = Idle;
            break;

        default:
            error = 1;
	}	
	return (error);
}

//*****************************************************************************
//Output logic routine 
//*****************************************************************************
char output_logic(void){
    char error = 0;
    switch (current_state) {
        case Idle:
            Timer_Out_value = 0;
            break;

        case Start_TMR0:
            Timer_Out_value = 1; //Here start the timing period
            TMR0_value = 248;//It has to be 250, but there is some software overhead
            TMR0_POST_Scaler = 0;
            TMR0_POST_Scaler_Max_Value = 1625; //(6.5 s)

            PSA 	= 0; 	// Prescaler in use
            PS2 	= 0; 	// PS(2..0) = "011" programs a 1:16 Prescale value
            PS1 	= 1;
            PS0 	= 1;

            TMR0 = (256 - TMR0_value); //Load the TMR0 (parallel LD)

            TMR0IE  = 1;  	// Enable TMR0 interrupts

            break;

        case Timing:
            Timer_Out_value = 1; //Do nothing while timing ...
            break;

        case Stop_TMR0:
            Timer_Out_value = 0;    // Turn off the timer out

            TMR0IE  = 0;  	// Disable TMR0 interrupts
            break;

        default:
            error = 1;
    }
    write_outputs();
    return (error);
}
/* **************************************************************************
   ISR: Interrupt Service Routine to attend the external INT
   "interrupt" is the key word here to make this ISR function special
   ***************************************************************************/
static void interrupt service_routine (void){
    GIE = 0;    //Disable interrupts while attending one of them ...

    if(TMR0IF) {	// Is it a Timer0 overflow?
        TMR0IF = 0;	// Clear interrupt flag
        TMR0 = 256 - TMR0_value; // reload the TMR0
        if (TMR0_POST_Scaler < TMR0_POST_Scaler_Max_Value ){
            TMR0_POST_Scaler++;
        }
        else {
            TMR0_Flag = 1;    // Set the software flag
            TMR0_POST_Scaler = 0;
        }
    }

    if(INTF) { // INT interrupt? (if so, the hardware flag bit is set)
              // Read the PIC16F877A datasheet, Chapter 14, section 11.
        Trigger_Flag = 1;   // Let's set our convenient "software" flag,
        INTF = 0;       // Clear interrupt "hardware" flag
    }                   // So that the system can be interrupted again
                        // for example to retrigger the timer

    GIE = 1;            //Enable general interrupts
}

/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables.
 Many of them are "char" to be monitored easily using the watch window.
  =========================================================================== */
void read_inputs (void){
    RT_value = (PORTA & 0b00010000)>>4;//Read the input and set the variable
}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting
	microcontroller variables into electrical signals.
   ========================================================================== */

// Using bit-wise operations you can write all the port in a single statement
  void write_outputs(void){
    PORTB = (PORTB & 0b11110111) | (Timer_Out_value << 3);
}
